#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include "usart.h"
#include "fifo.h" // erklrt im Artikel "FIFO mit avr-gcc"

#define BAUDRATE 9600

// FIFO-Objekte und Puffer fr die Ein- und Ausgabe 

#define BUFSIZE_IN  0x10
uint8_t inbuf[BUFSIZE_IN];
fifo_t infifo;

#define BUFSIZE_OUT 0x40
uint8_t outbuf[BUFSIZE_OUT];
fifo_t outfifo;

void uart_init()
{
    uint8_t sreg = SREG;
	LINBTR = (1<<LDISR) | 16;
    uint16_t ubrr = (uint16_t) ((uint32_t) F_CPU/((unsigned long)(LINBTR & 0x3F)*BAUDRATE) - 1);

    LINBRRH = (uint8_t) (ubrr>>8);
    LINBRRL = (uint8_t) (ubrr);

    // Interrupts kurz deaktivieren 
    cli();

    // Data mode 8N1, asynchron 
    LINCR = (1 << LENA);
    // UART Receiver und Transmitter anschalten
	LINCR |= ((1 << LCMD0) | (1 << LCMD1) | (1<<LCMD2));
	// Receive-Interrupt aktivieren
    LINENIR = (1 << LENRXOK);

	uart_flush_rec();

    // Rcksetzen von Receive und Transmit Complete-Flags 
    LINSIR = (1 << LRXOK) | (1 << LTXOK);

    // Global Interrupt-Flag wieder herstellen 
    SREG = sreg;

    // FIFOs fr Ein- und Ausgabe initialisieren 
    fifo_init (&infifo,   inbuf, BUFSIZE_IN);
    fifo_init (&outfifo, outbuf, BUFSIZE_OUT);
}

void uart_flush_rec()
{
    // Flush Receive-Buffer (entfernen evtl. vorhandener ungltiger Werte) 
    while (LINSIR & (1 << LRXOK))
    {
        // UDR auslesen (Wert wird nicht verwendet) 
        LINDAT;
    }
}

int uart_getc_nowait ()
{
    return fifo_get_nowait (&infifo);
}

uint8_t uart_getc_wait ()
{
    return fifo_get_wait (&infifo);
}

int uart_putc (const uint8_t c)
{
    int ret = fifo_put (&outfifo, c);

    // USART Data Register (Transmit buffer) Empty Interrupt aktivieren
	if( !(LINENIR & (1 << LENTXOK)) ) { // Wir sind nicht im Sendeprozess
		LINENIR |= (1 << LENTXOK);
		LINDAT = _inline_fifo_get (&outfifo);
	}

    return ret;
}

// Einen 0-terminierten String bertragen. 
void uart_puts (const char *s)
{
    while (*s)
    {
        uart_putc (*s);
		s++;
    }
}

// Einen 0-terminierten String bertragen. Warten, wenn Ausgabepuffer voll.
void uart_puts_wait (const char *s)
{
    while (*s)
    {
        while ( !uart_putc (*s) );
		s++;
    }
}

void uart_puts_p (PGM_P s)
{
    while (1)
    {
        unsigned char c = pgm_read_byte (s);
        s++;
        if (c == '\0')
            break;
        uart_putc (c);
    }
}

void uart_puts_p_wait (PGM_P s)
{
    while (1)
    {
        unsigned char c = pgm_read_byte (s);
        s++;
        if (c == '\0')
            break;
        while ( !uart_putc (c) );
    }
}




SIGNAL (LIN_TC_vect)
{
	if( LINSIR & (1<<LRXOK)) {
		// USART Receive Complete Interrupt
		// Empfangene Zeichen werden in die Eingangs-FIFO gespeichert und warten dort 
		_inline_fifo_put (&infifo, LINDAT);
	}
	if( LINSIR & (1<<LTXOK)) {
		// USART Data Register (Transmit buffer) Empty Interrupt
		// Ein Zeichen aus der Ausgabe-FIFO lesen und ausgeben 
		// Ist das Zeichen fertig ausgegeben, wird ein neuer SIG_USART_DATA-IRQ getriggert 
		// Ist die FIFO leer, deaktiviert die ISR ihren eigenen IRQ. 
		if (outfifo.count > 0)
		   LINDAT = _inline_fifo_get (&outfifo);
		else
			LINENIR &= ~(1 << LENTXOK);
	}
}
